/*eslint no-var: "error"*/
/*eslint no-console: "off"*/
/*eslint-env es6*/

import Config from "core/config";

/**
 * Start a new debugger
 * @param {*} handle The string to attach to all messages from this debugger
 * @returns Debugger object
 */
export default function (handle) {
    let output_enabled = Config.developerdebug;
    if(output_enabled){
        console.warn(`In development environment. Debugger output enabled for ${handle}`);
    } else {
        console.warn(`In production environment. Debugger output disabled for ${handle}`);
    }

    return {
        write() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.info.apply(console, args);
            }
        },
        info() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.info.apply(console, args);
            }
        },
        warn() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.warn.apply(console, args);
            }
        },
        error() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.error.apply(console, args);
            }
        },
        time() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.time.apply(console, args);
            }
        },
        timeEnd() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.timeEnd.apply(console, args);
            }
        },
        timeLog() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.timeLog.apply(console, args);
            }
        },
        timeStamp() {
            if (output_enabled) {
                let args = Array.prototype.slice.call(arguments);
                args.unshift(handle + ": ");
                console.timeStamp.apply(console, args);
            }
        },
        enable: function debugger_enable() {
            output_enabled = true;
        },

        disable: function debugger_disable() {
            output_enabled = false;
        }
    };

}
