<?php
// This file is part of the Studyplan plugin for Moodle
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.
/**
 * View study plans - teacher view and student view
 * @package    local_treestudyplan
 * @copyright  2023 P.M. Kuipers
 * @license    https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

require_once("../../config.php");

use \local_treestudyplan\premium;

require_once($CFG->libdir.'/weblib.php');

$systemcontext = context_system::instance();

$PAGE->set_url("/local/treestudyplan/coach.php", array());
require_login();


$PAGE->set_pagelayout('base');
$PAGE->set_context($systemcontext);
$PAGE->set_title(get_string('coaching_plans', 'local_treestudyplan'));
$PAGE->set_heading(get_string('coaching_plans', 'local_treestudyplan'));


premium::require_premium();
if ( ! (\get_config("local_treestudyplan","enablecoach")) ) {
    throw new \moodle_exception("error:coachdisabled","local_treestudyplan");
}
   

// Load javascripts and specific css.
$PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/bootstrap-vue/bootstrap-vue.css'));
if ($CFG->debugdeveloper) {
    $PAGE->requires->css(new moodle_url($CFG->wwwroot.'/local/treestudyplan/css/devstyles.css'));
}
$PAGE->requires->js_call_amd('local_treestudyplan/page-coach', 'init', []);

/**
 * Shortcut function to provide translations
 *
 * @param mixed $str Translation key
 * @param null|string[] $param Parameters to pass to translation
 * @param string $plugin Location to search for translation strings
 * @return string Translation of key
 */
function t($str, $param = null, $plugin = 'local_treestudyplan') {
    print get_string($str, $plugin, $param);
}

print $OUTPUT->header();
?>
<div id='root'>
    <div class='vue-loader' v-show='false'>
        <div class="spinner-border text-primary" role="status">
            <span class="sr-only">Loading...</span>
        </div>
    </div>
    <div v-cloak>
        <div class="m-buttonbar" style="margin-bottom: 1em;">
            <template v-if="displayedstudyplan">
                <a href='#' @click.prevent='closeStudyplan'
                    ><i style='font-size: 150%;' class='fa fa-chevron-left'></i> <?php t('back');?></a>
                <span><?php t("studyplan_select"); ?></span>&nbsp;
                <s-studyplan-details 
                    v-model="displayedstudyplan" 
                    v-if="displayedstudyplan.description"
                ></s-studyplan-details>
                <div class="flex-grow-1"><!-- Spacer to align student selector right --></div> 
                <span><?php t('selectstudent_btn') ?></span>
                <s-prevnext-selector
                    :options="associatedstudents"
                    title="firstname"
                    v-model="selectedstudent"
                    defaultselectable
                    grouped
                    optionsfield='users'
                    arrows
                    @change="showStudentView"
                    class="ml-2"
                    variant="primary"
                    >
                    <template v-slot="{value}">{{value.firstname}} {{value.lastname}}</template>
                    <template #defaultlabel><span class='text-primary'><?php t("coacheditmode"); ?></span></template>
                </s-prevnext-selector>
            </template>
        </div>
        <div class='t-studyplan-container'>
            <h2 v-if='displayedstudyplan&& selectedstudent'
                >{{selectedstudent.firstname}} {{selectedstudent.lastname}} - {{displayedstudyplan.name}}</h2>
            <h2 v-else-if='displayedstudyplan'><?php t("showoverview"); ?> - {{displayedstudyplan.name}}</h2>
            <template v-if="!loadingstudyplan && displayedstudyplan">
                <r-studyplan v-if="selectedstudent"
                         v-model='displayedstudyplan'
                         coaching
                ></r-studyplan>
                <t-studyplan
                    v-else
                    v-model='displayedstudyplan'
                    coaching
                ></t-studyplan>
            </template>
            <div v-else-if='loadingstudyplan' class="spinner-border text-primary" role="status">
                <span class="sr-only">Loading...</span>
            </div>
            <div v-else class='t-studyplan-notselected'>
                <p><?php t("studyplan_noneselected"); ?></p>
                <b-card-group deck>
                    <s-studyplan-card
                        v-for='(studyplan, planindex) in studyplans'
                        :key='studyplan.id'
                        v-model='studyplans[planindex]'
                        open
                        @open='selectStudyplan(studyplan)'
                        ></s-studyplan-card>
                </b-card-group>
            </div>
        </div>
    </div>
</div>
<?php

print $OUTPUT->footer();
